/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_accept_setup.h"
#include "setup/exs_bind_setup.h"
#include "setup/exs_close_setup.h"
#include "setup/exs_connect_setup.h"
#include "setup/exs_init_setup.h"
#include "setup/exs_listen_setup.h"
#include "setup/exs_qcreate_setup.h"
#include "setup/exs_qdequeue_setup.h"
#include "setup/exs_recv_setup.h"
#include "setup/exs_send_setup.h"
#include "setup/exs_socket_setup.h"

#define RECV_TESTS 10

static char *all_recv_tests[RECV_TESTS] = {
	"exs_recv(fd, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh) [no exs_init]",
	"exs_recv(BAD_FD, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)",
	"exs_recv(NOT_CONNECTED_FD, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)",
	"exs_recv(fd, BAD_BUFFER, RECV_BUFSIZE, flags, qh, ah, mh)",
	"exs_recv(fd, recv_buf, ZERO_LENGTH, flags, qh, ah, mh) [length < 1 ]",
	"exs_recv(fd, recv_buf, MAX_LENGTH+1, flags, qh, ah, mh) [length > RDMA channel]",
	"exs_recv(fd, recv_buf, RECV_BUFSIZE, BAD_FLAG, qh, ah, mh)",
	"exs_recv(fd, recv_buf, RECV_BUFSIZE, flags, INVALID_QHANDLE, ah, mh)",
	"exs_recv(fd, recv_buf, RECV_BUFSIZE, 0, INVALID_QHANDLE, ah, mh)",
	"exs_recv(fd, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)",
};

static void *test_recv(void *test_void_ptr)
{
	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		int ret =
		    exs_recv(p->recv_fd, p->recv_buf, p->RECV_BUFSIZE,
			     p->flags, p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* EBADF: bad fd */
	else if (test == 2) {
		p->test_value = "EBADF";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (p->mode > 0)
			if (ok_exs_qcreate(p) < 0)
				exit(EXIT_FAILURE);
		int ret =
		    exs_recv(-1, p->recv_buf, p->RECV_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EBADF);
	}

	/* ENOTCONN: socket was never connected */
	else if (test == 3) {
		p->test_value = "ENOTCONN";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		if (p->mode > 0) {
			if (ok_exs_qcreate(p) < 0)
				exit(EXIT_FAILURE);
		}
		int ret =
		    exs_recv(p->fd, p->recv_buf, p->RECV_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    ENOTCONN);
	}

	/* EINVAL: buffer is null */
	else if (test == 4) {	//EINVAL
		p->test_value = "EINVAL";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_recv(p->recv_fd, NULL, p->RECV_BUFSIZE, p->flags,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: length is not at least 1 */
	else if (test == 5) {
		p->test_value = "EINVAL";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_recv(p->recv_fd, p->recv_buf, 0, p->flags, p->qh,
			     p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: length is greater than RDMA channel */
	else if (test == 6) {
		p->test_value = "EINVAL";
		p->test_name = all_recv_tests[test - 1];
		pid_t pid = fork();
		if (pid == 0) {
			/** blocking client **/
			if (p->mode == 0) {
				int fd = open(exs_test_fifo_1, O_RDONLY);

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);

				close(fd);

				if (ok_exs_send(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_WRONLY);

				close(fd);
			}
			/** non-blocking client **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);

				if (ok_exs_send(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			}
			exit(EXIT_SUCCESS);
		} else {
			setup_fault_catcher(p, pid);
			/** blocking server **/
			if (p->mode == 0) {

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);

				int fd = open(exs_test_fifo_1, O_WRONLY);

				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				int ret =
				    exs_recv(p->recv_fd, p->recv_buf,
					     1073741825, p->flags, p->qh,
					     p->ah, p->mh);

				print_macro(test, p->test_value,
					    p->test_name, ret, -1, EINVAL);

				close(fd);

				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_RDONLY);

				close(fd);
			}
			/** non-blocking server **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
				int ret =
				    exs_recv(p->recv_fd, p->recv_buf,
					     1073741825, p->flags, p->qh,
					     p->ah, p->mh);

				print_macro(test, p->test_value,
					    p->test_name, ret, -1, EINVAL);
			}
		}
	}

	/* EOPNOTSUPP: flags are not supported */
	else if (test == 7) {
		p->test_value = "EOPNOTSUPP";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_recv(p->recv_fd, p->recv_buf, p->RECV_BUFSIZE, 2,
			     p->qh, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EOPNOTSUPP);
	}

	/* EINVAL: qhandle is invalid */
	else if (test == 8) {
		p->test_value = "EINVAL";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_recv(p->recv_fd, p->recv_buf, p->RECV_BUFSIZE,
			     p->flags, (exs_qhandle_t) - 1, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: qhandle is invalid (NULL) */
	else if (test == 9) {
		p->test_value = "EINVAL";
		p->test_name = all_recv_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_recv(p->recv_fd, p->recv_buf, p->RECV_BUFSIZE, 0,
			     NULL, p->ah, p->mh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* OK_EXS_RECV */
	else if (test == 10) {
		p->test_value = "OK_EXS_RECV";
		p->test_name = all_recv_tests[test - 1];
		pid_t pid = fork();
		if (pid == 0) {
			/** blocking client **/
			if (p->mode == 0) {
				int fd = open(exs_test_fifo_1, O_RDONLY);

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);

				close(fd);

				if (ok_exs_send(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_WRONLY);

				close(fd);
			}
			/** non-blocking client **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_connect(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);

				if (ok_exs_send(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			}
			exit(EXIT_SUCCESS);
		} else {
			setup_fault_catcher(p, pid);
			/** blocking server **/
			if (p->mode == 0) {

				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);

				int fd = open(exs_test_fifo_1, O_WRONLY);

				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				int ret = ok_exs_recv(p);

				if (ret < p->RECV_BUFSIZE) {
					print_macro(test, p->test_value,
						    p->test_name, ret, -1,
						    ENOTSUP);
				} else {
					print_macro(test, p->test_value,
						    p->test_name, ret,
						    p->RECV_BUFSIZE, 0);
				}

				close(fd);

				if (ok_exs_close(p) < 0)
					exit(EXIT_FAILURE);

				fd = open(exs_test_fifo_2, O_RDONLY);

				close(fd);
			}
			/** non-blocking server **/
			if (p->mode > 0) {
				if (ok_exs_init(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_socket(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_bind(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_listen(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_accept(p) < 0)
					exit(EXIT_FAILURE);
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
				int ret = ok_exs_recv(p);

				if (ret < 0) {
					print_macro(test, p->test_value,
						    p->test_name, ret, -1,
						    ENOTSUP);
				} else {
					print_macro(test, p->test_value,
						    p->test_name, ret, 0,
						    0);
				}
			}
		}
	}

    /** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}

	return NULL;
}

int main(int argc, char *argv[])
{
	execute(argc, argv, test_recv, all_recv_tests, RECV_TESTS);
}
