/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_write_setup.h"
#include "setup/exs_server_setup.h"

#define WRITE_TESTS 11

static char *all_write_tests[WRITE_TESTS] = {
	"exs_write(fd, send_buf, SEND_BUFSIZE) [no exs_init]",
	"exs_write(NOT_CONNECTED_FD, send_buf, SEND_BUFSIZE)",
	"exs_write(fd, NULL_BUFFER, SEND_BUFSIZE)",
	"",
	"exs_write(fd, send_buf, SEND_BUFSIZE) [gracefull shut down for writing]",
	"",
	"",
	"",
	"",
	"",
	"exs_write(fd, send_buf, SEND_BUFSIZE)"
};

static void *test_write(void *test_void_ptr)
{
	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	if (p->mode > 0)
		return 0;

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_write_tests[test - 1];
		setup_fault_catcher(p, 0);
		int ret = exs_write(p->fd, p->send_buf, p->SEND_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* ENOTCONN: socket never connected */
	else if (test == 2) {
		p->test_value = "ENOTCONN";
		p->test_name = all_write_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		int ret = exs_write(p->fd, p->send_buf, p->SEND_BUFSIZE);

		print_macro(test, p->test_value, p->test_name, ret, -1,
			    ENOTCONN);
	}

	/* EINVAL: buffer is null */
	else if (test == 3) {
		p->test_value = "EINVAL";
		p->test_name = all_write_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		int ret = exs_write(p->fd, NULL, p->SEND_BUFSIZE);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EBUSY: no credits available */
	else if (test == 4) {	//TODO
	}

	/* EPIPE: gracefull shut down for writing */
	else if (test == 5) {
		p->test_value = "EPIPE";
		p->test_name = all_write_tests[test - 1];

		pid_t pid = fork();
		if (pid == 0) {
			exs_server(p);
			exit(EXIT_SUCCESS);
		} else {
			setup_fault_catcher(p, pid);
			int fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			close(fd);

			int shut_ret =
			    exs_shutdown(p->fd, SHUT_WR, p->flags, p->qh,
					 p->ah);
			if (shut_ret < 0) {
				print_error("exs_shutdown");
				exit(EXIT_FAILURE);
			}

			/** start exs_send **/

			int ret =
			    exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE,
				     p->flags, p->qh, p->ah, p->mh);
			print_macro(test, p->test_value, p->test_name, ret,
				    -1, EPIPE);

			/** end send **/

			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_WRONLY);

			close(fd);
		}
	} else if (test == 6) {	/*ECONNRESET: abruptly terminated */

	} else if (test == 7) {	/*ECANCELED: operation has been cancelled */

	} else if (test == 8) {	/*ECONNRESET: forcibly closed by a peer //TODO */

	} else if (test == 9) {	/*EFAULT: detected address was not vaild for buffer argument */

	} else if (test == 10) {	/*ENETDOWN: local interface is down */

	}

	/* OK_EXS_WRITE */
	else if (test == 11) {
		p->test_value = "OK_EXS_WRITE";
		p->test_name = all_write_tests[test - 1];
		pid_t pid = fork();
		if (pid == 0) {
			exs_server(p);
			exit(EXIT_SUCCESS);
		} else {
			setup_fault_catcher(p, pid);
			int fd = open(exs_test_fifo_1, O_RDONLY);

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_connect(p) < 0)
				exit(EXIT_FAILURE);

			close(fd);

			/** start exs_write **/

			int ret = ok_exs_write(p);

			if (ret < p->SEND_BUFSIZE) {
				print_macro(test, p->test_value,
					    p->test_name, ret, -1,
					    ENOTSUP);
			} else {
				print_macro(test, p->test_value,
					    p->test_name, ret,
					    p->SEND_BUFSIZE, 0);
			}

			/** end send **/

			if (ok_exs_close(p) < 0)
				exit(EXIT_FAILURE);

			fd = open(exs_test_fifo_2, O_WRONLY);

			close(fd);
		}
	}

    /** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}

	return NULL;
}

int main(int argc, char *argv[])
{
	execute(argc, argv, test_write, all_write_tests, WRITE_TESTS);
}
